import argparse
import sys
import subprocess
import platform
import os
from typing import List, Dict

def get_system_info():
    """Retrieves system information (RAM, Chip) on macOS."""
    info = {}
    
    # Get Chip Name
    try:
        result = subprocess.run(["sysctl", "-n", "machdep.cpu.brand_string"], capture_output=True, text=True)
        info["chip"] = result.stdout.strip()
    except Exception:
        info["chip"] = "Unknown Apple Silicon"

    # Get RAM
    try:
        result = subprocess.run(["sysctl", "-n", "hw.memsize"], capture_output=True, text=True)
        bytes_ram = int(result.stdout.strip())
        gb_ram = bytes_ram / (1024**3)
        info["ram"] = f"{gb_ram:.1f} GB"
        info["ram_bytes"] = bytes_ram
    except Exception:
        info["ram"] = "Unknown"
        info["ram_bytes"] = 0

    info["os"] = f"{platform.system()} {platform.release()}"
    return info

def recommend_models(ram_bytes: int) -> List[Dict[str, str]]:
    """Recommends models based on available RAM."""
    # Rough estimates for 4-bit quantized models
    # 7B ~ 4-5GB
    # 13B ~ 8-9GB
    # 70B ~ 40GB
    
    models = [
        {"name": "TinyLlama-1.1B-Chat-v1.0", "id": "TinyLlama/TinyLlama-1.1B-Chat-v1.0", "size_gb": 1.0},
        {"name": "Llama-2-7b-chat", "id": "meta-llama/Llama-2-7b-chat-hf", "size_gb": 5.0},
        {"name": "Mistral-7B-Instruct-v0.2", "id": "mistralai/Mistral-7B-Instruct-v0.2", "size_gb": 5.0},
        {"name": "Llama-2-13b-chat", "id": "meta-llama/Llama-2-13b-chat-hf", "size_gb": 9.0},
        {"name": "Mixtral-8x7B-Instruct-v0.1", "id": "mistralai/Mixtral-8x7B-Instruct-v0.1", "size_gb": 26.0},
    ]
    
    # Filter based on RAM (leave 2GB for system)
    available_ram_gb = ram_bytes / (1024**3)
    safe_limit = available_ram_gb - 2.0
    
    recommended = []
    for m in models:
        status = "✅ Recommended" if m["size_gb"] <= safe_limit else "⚠️  May swap/crash"
        if m["size_gb"] > available_ram_gb:
            status = "❌ Insufficient RAM"
        
        m["status"] = status
        recommended.append(m)
        
    return recommended

def download_model_interactive():
    print("\n--- System Check ---")
    sys_info = get_system_info()
    print(f"Chip: {sys_info['chip']}")
    print(f"RAM:  {sys_info['ram']}")
    
    print("\n--- Available Models ---")
    models = recommend_models(sys_info.get("ram_bytes", 0))
    
    for idx, m in enumerate(models):
        print(f"{idx + 1}. {m['name']:<30} | ~{m['size_gb']} GB | {m['status']}")
    
    print("\nSelect a model to download (enter number) or 'q' to quit:")
    choice = input("> ").strip()
    
    if choice.lower() == 'q':
        return
    
    try:
        idx = int(choice) - 1
        if 0 <= idx < len(models):
            selected = models[idx]
            print(f"\nDownloading {selected['name']} from Hugging Face...")
            print(f"Repo ID: {selected['id']}")
            
            # Use huggingface_hub to download
            try:
                from huggingface_hub import snapshot_download
                local_dir = f"models/{selected['name']}"
                print(f"Saving to {local_dir}...")
                snapshot_download(repo_id=selected['id'], local_dir=local_dir, local_dir_use_symlinks=False)
                print(f"\n✅ Download complete! You can run it with:\n orchard run --model {local_dir}")
            except ImportError:
                print("Error: 'huggingface_hub' is not installed. Please install it via pip.")
                print("pip install huggingface_hub")
            except Exception as e:
                print(f"Error downloading: {e}")
        else:
            print("Invalid selection.")
    except ValueError:
        print("Invalid input.")

def main():
    parser = argparse.ArgumentParser(description="Orchard CLI - Apple Silicon LLM Runtime")
    subparsers = parser.add_subparsers(dest="command", help="Command to run")
    
    # Info command
    subparsers.add_parser("info", help="Show system information")
    
    # Download command
    subparsers.add_parser("download", help="Download a model")
    
    # Run command
    run_parser = subparsers.add_parser("run", help="Run a model")
    run_parser.add_argument("--model", type=str, required=True, help="Path to model directory")
    run_parser.add_argument("--prompt", type=str, default="Hello, how are you?", help="Initial prompt")
    
    # Optimize command
    optimize_parser = subparsers.add_parser("optimize", help="Optimize (quantize) a model for faster loading")
    optimize_parser.add_argument("--model", type=str, required=True, help="Path to source model directory")
    optimize_parser.add_argument("--output", type=str, required=True, help="Path to output directory")

    args = parser.parse_args()
    
    if args.command == "info":
        info = get_system_info()
        print(f"Orchard Runtime v0.1.0")
        print(f"----------------------")
        print(f"OS:   {info['os']}")
        print(f"Chip: {info['chip']}")
        print(f"RAM:  {info['ram']}")
        
    elif args.command == "download":
        download_model_interactive()
        
    elif args.command == "optimize":
        try:
            from orchard.optimizer import optimize_model
            optimize_model(args.model, args.output)
        except Exception as e:
            print(f"Error optimizing model: {e}")
            import traceback
            traceback.print_exc()

    elif args.command == "run":
        print(f"Loading model from {args.model}...")
        try:
            from orchard.model import Llama
            model = Llama(args.model)
            print(f"\nGenerating response to: '{args.prompt}'\n")
            results = model.generate_batch([args.prompt])
            print(results[0])
        except Exception as e:
            print(f"Error running model: {e}")
            import traceback
            traceback.print_exc()
        
    else:
        parser.print_help()

if __name__ == "__main__":
    main()



from .model import Llama
from .loader import load_weights
import os
import orchard_core

# Monkey patch initialize to automatically pass the kernel path
_original_initialize = orchard_core.MetalBackend.initialize

def _initialize_wrapper(self, resource_path=None):
    if resource_path is None:
        # Default to the installed kernels directory
        resource_path = os.path.join(os.path.dirname(__file__), "kernels")
    return _original_initialize(self, resource_path)

orchard_core.MetalBackend.initialize = _initialize_wrapper

__version__ = "0.1.0"



import numpy as np

def quantize_q4_0(weight_fp32, group_size=32):
    """
    Quantize a weight matrix (K, N) or (N, K) to Q4_0 format.
    Returns (packed_weights, scales).
    
    Args:
        weight_fp32: Numpy array of shape (out_features, in_features)
        group_size: Block size for quantization (default 32)
    """
    # Ensure shape is (N, K) where K is input dimension (reduction dim)
    # Our kernels expect row-major weights where each row is a output neuron.
    # weight_fp32 shape: [N, K]
    
    N, K = weight_fp32.shape
    assert K % group_size == 0, f"Input dimension {K} must be divisible by group size {group_size}"
    
    # Reshape to (N, K/group_size, group_size) to process blocks
    # We want to find scale for each block of 32 weights.
    w_reshaped = weight_fp32.reshape(N, K // group_size, group_size)
    
    # Find max absolute value in each block
    max_abs = np.max(np.abs(w_reshaped), axis=2)
    
    # Calculate scales (fp16)
    # We map range [-max, max] to [-7, 7] (integers)
    # scale = max_abs / 7.0
    scales = max_abs / 7.0
    scales = scales.astype(np.float16)
    
    # Avoid divide by zero
    scales[scales == 0] = 1e-5
    
    # Quantize
    # w_quant = round(w / scale) + 8
    # Result is in [1, 15] range usually. 0 is reserved or just unused in this symmetric scheme?
    # Actually Q4_0 in llama.cpp usually maps -8 to 7 or similar.
    # Let's stick to the kernel implementation:
    # float val = (float(packed & 0xF) - 8.0f) * scale;
    # So value 8 -> 0.0.
    # Value 0 -> -8.0 * scale.
    # Value 15 -> 7.0 * scale.
    
    # Broadcast scales back
    scales_expanded = scales[:, :, np.newaxis]
    
    # Divide and round
    w_scaled = w_reshaped / scales_expanded
    w_quant = np.round(w_scaled + 8.0).astype(np.int8)
    
    # Clamp to [0, 15]
    w_quant = np.clip(w_quant, 0, 15).astype(np.uint8)
    
    # Pack 2 weights into 1 byte
    # Layout: Low nibble = even index, High nibble = odd index?
    # Kernel: 
    # uchar4 lo = packed & 0x0F; -> w0
    # uchar4 hi = packed >> 4;   -> w1
    # So low bits are the first weight.
    
    # We need to pack along the last dimension (group_size)
    # w_quant shape: (N, blocks, 32)
    # We want to pack pairs (0,1), (2,3), etc.
    
    # Reshape to (N, blocks, 16, 2)
    w_pairs = w_quant.reshape(N, K // group_size, group_size // 2, 2)
    
    # w0 is w_pairs[..., 0], w1 is w_pairs[..., 1]
    # packed = (w1 << 4) | w0
    low = w_pairs[..., 0]
    high = w_pairs[..., 1]
    packed = (high << 4) | low
    
    # Result shape: (N, blocks, 16) -> flatten to (N, K/2)
    packed_weights = packed.reshape(N, K // 2)
    
    return packed_weights.astype(np.uint8), scales



import os
import sys
import unittest

# Add root to path
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

class TestOrchardStructure(unittest.TestCase):
    def test_kernels_exist(self):
        """Check if kernel files were moved correctly."""
        import orchard
        orchard_path = os.path.dirname(orchard.__file__)
        kernels_path = os.path.join(orchard_path, "kernels")
        self.assertTrue(os.path.exists(kernels_path))
        self.assertTrue(os.path.exists(os.path.join(kernels_path, "matmul.metal")))
        self.assertTrue(os.path.exists(os.path.join(kernels_path, "gemm_int4.metal")))

    def test_cli_import(self):
        """Check if CLI module is importable."""
        from orchard import cli
        self.assertTrue(hasattr(cli, "main"))

if __name__ == "__main__":
    unittest.main()



import orchard_core
import numpy as np
import os
import time
from typing import List, Optional, Tuple

def precompute_freqs_cis(dim: int, end: int, theta: float = 10000.0):
    freqs = 1.0 / (theta ** (np.arange(0, dim, 2)[: (dim // 2)].astype(np.float32) / dim))
    t = np.arange(end).astype(np.float32)
    freqs = np.outer(t, freqs)
    freqs_cos = np.cos(freqs)
    freqs_sin = np.sin(freqs)
    return freqs_cos, freqs_sin

class LoraAdapter:
    def __init__(self, backend, r, alpha, scaling):
        self.backend = backend
        self.r = r
        self.alpha = alpha
        self.scaling = scaling
        # Stored in format suitable for matmul:
        # A: [in_dim, r]
        # B: [r, out_dim]
        self.lora_A = None 
        self.lora_B = None

class LlamaLayer:
    def __init__(self, backend, config, layer_idx):
        self.backend = backend
        self.config = config
        self.layer_idx = layer_idx
        self.hidden_dim = config['hidden_size']
        self.heads = config['num_attention_heads']
        self.head_dim = self.hidden_dim // self.heads
        self.intermediate_dim = config['intermediate_size']
        
        # Weights (Quantized)
        self.wq_w = None; self.wq_s = None
        self.wk_w = None; self.wk_s = None
        self.wv_w = None; self.wv_s = None
        self.wo_w = None; self.wo_s = None
        
        self.w1_w = None; self.w1_s = None # Gate
        self.w2_w = None; self.w2_s = None # Down
        self.w3_w = None; self.w3_s = None # Up
        
        self.attn_norm = None
        self.ffn_norm = None
        
        # LoRA Adapters (key: module name, value: LoraAdapter)
        # Modules: wq, wk, wv, wo, w1, w2, w3
        self.lora_adapters = {}
        
        # KV Cache (CPU numpy arrays, uploaded on demand)
        # List of (Heads, HeadDim, SeqLen) per sequence
        # We assume a max batch size of 32 for now
        self.k_cache_np = [np.zeros((self.heads, self.head_dim, 0), dtype=np.float32) for _ in range(32)]
        self.v_cache_np = [np.zeros((self.heads, 0, self.head_dim), dtype=np.float32) for _ in range(32)]

    def _apply_lora(self, x, output, module_name, batch_size, in_dim, out_dim):
        if module_name in self.lora_adapters:
            adapter = self.lora_adapters[module_name]
            # x: [Batch, In]
            # A: [In, r]
            # B: [r, Out]
            
            # Temp = x @ A
            temp = orchard_core.Tensor(self.backend, [batch_size * adapter.r], orchard_core.DType.Float32)
            self.backend.run_matmul(x, adapter.lora_A, temp, batch_size, adapter.r, in_dim)
            
            # Result = Temp @ B
            lora_out = orchard_core.Tensor(self.backend, [batch_size * out_dim], orchard_core.DType.Float32)
            self.backend.run_matmul(temp, adapter.lora_B, lora_out, batch_size, out_dim, adapter.r)
            
            # Add to output (B is pre-scaled)
            self.backend.run_add(output, lora_out, output, batch_size * out_dim)

    def forward(self, x, freqs_cos, freqs_sin, batch_indices):
        # x: [Batch, Hidden] (Tensor)
        # freqs_cos/sin: [Batch, HeadDim/2] (Tensor) - RoPE for current positions
        # batch_indices: List[int] - Indices into the KV cache for each item in the batch
        
        batch_size = len(batch_indices)
        
        # 1. RMSNorm
        x_norm = orchard_core.Tensor(self.backend, [batch_size * self.hidden_dim], orchard_core.DType.Float32)
        self.backend.run_rmsnorm(x, self.attn_norm, x_norm, self.config['rms_norm_eps'], self.hidden_dim, batch_size)
        
        # 2. Q, K, V Projections
        q = orchard_core.Tensor(self.backend, [batch_size * self.hidden_dim], orchard_core.DType.Float32)
        k = orchard_core.Tensor(self.backend, [batch_size * self.hidden_dim], orchard_core.DType.Float32)
        v = orchard_core.Tensor(self.backend, [batch_size * self.hidden_dim], orchard_core.DType.Float32)
        
        if batch_size == 1:
            self.backend.run_gemv_q4_0(self.wq_w, self.wq_s, x_norm, q, self.hidden_dim, self.hidden_dim)
            self.backend.run_gemv_q4_0(self.wk_w, self.wk_s, x_norm, k, self.hidden_dim, self.hidden_dim)
            self.backend.run_gemv_q4_0(self.wv_w, self.wv_s, x_norm, v, self.hidden_dim, self.hidden_dim)
        else:
            self.backend.run_gemm_q4_0(self.wq_w, self.wq_s, x_norm, q, self.hidden_dim, self.hidden_dim, batch_size)
            self.backend.run_gemm_q4_0(self.wk_w, self.wk_s, x_norm, k, self.hidden_dim, self.hidden_dim, batch_size)
            self.backend.run_gemm_q4_0(self.wv_w, self.wv_s, x_norm, v, self.hidden_dim, self.hidden_dim, batch_size)
            
        # Apply LoRA
        self._apply_lora(x_norm, q, "wq", batch_size, self.hidden_dim, self.hidden_dim)
        self._apply_lora(x_norm, k, "wk", batch_size, self.hidden_dim, self.hidden_dim)
        self._apply_lora(x_norm, v, "wv", batch_size, self.hidden_dim, self.hidden_dim)
        
        # 3. RoPE
        self.backend.run_rope(q, freqs_cos, freqs_sin, q, self.head_dim, self.heads, batch_size)
        self.backend.run_rope(k, freqs_cos, freqs_sin, k, self.head_dim, self.heads, batch_size)
        
        # 4. Update KV Cache & Attention
        q_cpu = np.array(q.copy_to_host(), copy=False).reshape(batch_size, self.heads, self.head_dim)
        k_cpu = np.array(k.copy_to_host(), copy=False).reshape(batch_size, self.heads, self.head_dim)
        v_cpu = np.array(v.copy_to_host(), copy=False).reshape(batch_size, self.heads, self.head_dim)
        
        output_cpu = np.zeros((batch_size, self.heads, self.head_dim), dtype=np.float32)
        scale = 1.0 / np.sqrt(self.head_dim)
        
        for i, b_idx in enumerate(batch_indices):
            # Update Cache for this sequence (b_idx)
            # k_cpu[i]: (Heads, HeadDim)
            self.k_cache_np[b_idx] = np.dstack([self.k_cache_np[b_idx], k_cpu[i][:, :, None]])
            self.v_cache_np[b_idx] = np.hstack([self.v_cache_np[b_idx], v_cpu[i][:, None, :]])
            
            # Attention (Per Head)
            for h in range(self.heads):
                qs = q_cpu[i, h] # (HeadDim)
                ks = self.k_cache_np[b_idx][h] # (HeadDim, Seq)
                vs = self.v_cache_np[b_idx][h] # (Seq, HeadDim)
                
                score = np.matmul(qs, ks) * scale
                score_max = np.max(score)
                exp_score = np.exp(score - score_max)
                softmax = exp_score / np.sum(exp_score)
                
                output_cpu[i, h] = np.matmul(softmax, vs)
            
        # Flatten output
        output_flat = output_cpu.flatten()
        
        # Upload back to GPU
        attn_output = orchard_core.Tensor(self.backend, [batch_size * self.hidden_dim], orchard_core.DType.Float32)
        attn_output.copy_from_host(output_flat)
        
        # 5. Output Projection
        attn_proj = orchard_core.Tensor(self.backend, [batch_size * self.hidden_dim], orchard_core.DType.Float32)
        if batch_size == 1:
            self.backend.run_gemv_q4_0(self.wo_w, self.wo_s, attn_output, attn_proj, self.hidden_dim, self.hidden_dim)
        else:
            self.backend.run_gemm_q4_0(self.wo_w, self.wo_s, attn_output, attn_proj, self.hidden_dim, self.hidden_dim, batch_size)
            
        # Apply LoRA
        self._apply_lora(attn_output, attn_proj, "wo", batch_size, self.hidden_dim, self.hidden_dim)
        
        # 6. Residual Add
        self.backend.run_add(x, attn_proj, x, batch_size * self.hidden_dim)
        
        # 7. FFN
        ffn_input = orchard_core.Tensor(self.backend, [batch_size * self.hidden_dim], orchard_core.DType.Float32)
        self.backend.run_rmsnorm(x, self.ffn_norm, ffn_input, self.config['rms_norm_eps'], self.hidden_dim, batch_size)
        
        # Gate & Up
        gate = orchard_core.Tensor(self.backend, [batch_size * self.intermediate_dim], orchard_core.DType.Float32)
        up = orchard_core.Tensor(self.backend, [batch_size * self.intermediate_dim], orchard_core.DType.Float32)
        
        if batch_size == 1:
            self.backend.run_gemv_q4_0(self.w1_w, self.w1_s, ffn_input, gate, self.hidden_dim, self.intermediate_dim)
            self.backend.run_gemv_q4_0(self.w3_w, self.w3_s, ffn_input, up, self.hidden_dim, self.intermediate_dim)
        else:
            self.backend.run_gemm_q4_0(self.w1_w, self.w1_s, ffn_input, gate, self.hidden_dim, self.intermediate_dim, batch_size)
            self.backend.run_gemm_q4_0(self.w3_w, self.w3_s, ffn_input, up, self.hidden_dim, self.intermediate_dim, batch_size)
            
        # Apply LoRA
        self._apply_lora(ffn_input, gate, "w1", batch_size, self.hidden_dim, self.intermediate_dim)
        self._apply_lora(ffn_input, up, "w3", batch_size, self.hidden_dim, self.intermediate_dim)
        
        # Silu & Mul
        self.backend.run_silu(gate, gate, batch_size * self.intermediate_dim)
        self.backend.run_mul(gate, up, gate, batch_size * self.intermediate_dim)
        
        # Down
        ffn_output = orchard_core.Tensor(self.backend, [batch_size * self.hidden_dim], orchard_core.DType.Float32)
        if batch_size == 1:
            self.backend.run_gemv_q4_0(self.w2_w, self.w2_s, gate, ffn_output, self.intermediate_dim, self.hidden_dim)
        else:
            self.backend.run_gemm_q4_0(self.w2_w, self.w2_s, gate, ffn_output, self.intermediate_dim, self.hidden_dim, batch_size)
            
        # Apply LoRA
        self._apply_lora(gate, ffn_output, "w2", batch_size, self.intermediate_dim, self.hidden_dim)
        
        # Residual Add
        self.backend.run_add(x, ffn_output, x, batch_size * self.hidden_dim)
        
        return x

class Llama:
    def __init__(self, model_path: str):
        self.backend = orchard_core.MetalBackend()
        
        # Find kernels path
        import os
        current_dir = os.path.dirname(os.path.abspath(__file__))
        kernels_path = os.path.join(current_dir, "kernels")
        
        self.backend.initialize(kernels_path)
        if not self.backend.is_available():
            raise RuntimeError("Metal backend not available")
            
        print(f"Orchard initialized on {self.backend.get_device_name()}")
        
        self.layers: List[LlamaLayer] = []
        self.config = {}
        self.load_model(model_path)
        
        # Precompute RoPE
        self.freqs_cos, self.freqs_sin = precompute_freqs_cis(
            self.config['hidden_size'] // self.config['num_attention_heads'],
            self.config.get('max_position_embeddings', 2048)
        )
        
        # Load Tokenizer
        try:
            from tokenizers import Tokenizer
            tokenizer_path = os.path.join(model_path, "tokenizer.json")
            if os.path.exists(tokenizer_path):
                self.tokenizer = Tokenizer.from_file(tokenizer_path)
            else:
                # Fallback or error
                print("Warning: tokenizer.json not found. Generation will fail if not provided manually.")
                self.tokenizer = None
        except ImportError:
            print("Warning: 'tokenizers' library not found. Install it with `pip install tokenizers`.")
            self.tokenizer = None
        
    def load_model(self, model_path: str):
        from .loader import load_weights
        load_weights(self, model_path)
        
    def load_lora(self, adapter_path: str):
        from .loader import load_lora_adapter
        load_lora_adapter(self, adapter_path)
        
    def _sample(self, logits, temperature=0.7):
        if temperature == 0:
            return np.argmax(logits)
        else:
            # Softmax
            # Shift logits for stability
            logits = logits - np.max(logits)
            probs = np.exp(logits / temperature) / np.sum(np.exp(logits / temperature))
            return np.random.choice(len(probs), p=probs)

    def generate_batch(self, prompts: List[str], max_tokens: int = 50, temperature: float = 0.7):
        if self.tokenizer is None:
            raise RuntimeError("Tokenizer not available")
            
        batch_size = len(prompts)
        
        # Tokenize
        batch_ids = [self.tokenizer.encode(p).ids for p in prompts]
        
        # Current position for each sequence
        # We need to track this because sequences might have different lengths
        # But for v0.1 we assume we process them in lockstep or just track pos.
        
        # Prefill
        # We process each prompt token by token to fill the cache.
        # Ideally we'd do this in parallel too, but for now let's just fill it up.
        
        print(f"Prefilling {batch_size} prompts...")
        current_tokens = []
        
        # We need to reset cache for these slots?
        # Assuming slots 0..batch_size-1 are free or we overwrite them.
        # The cache is just a list of numpy arrays, so we can just reset them.
        for b in range(batch_size):
            self.layers[0].k_cache_np[b] = np.zeros((self.layers[0].heads, self.layers[0].head_dim, 0), dtype=np.float32)
            self.layers[0].v_cache_np[b] = np.zeros((self.layers[0].heads, 0, self.layers[0].head_dim), dtype=np.float32)
            # We need to do this for all layers!
            for layer in self.layers:
                layer.k_cache_np[b] = np.zeros((layer.heads, layer.head_dim, 0), dtype=np.float32)
                layer.v_cache_np[b] = np.zeros((layer.heads, 0, layer.head_dim), dtype=np.float32)

        for b in range(batch_size):
            ids = batch_ids[b]
            for i, token_id in enumerate(ids):
                # Prepare input
                x_emb = self.embed_tokens[token_id]
                x = orchard_core.Tensor(self.backend, [self.config['hidden_size']], orchard_core.DType.Float32)
                x.copy_from_host(x_emb)
                
                # RoPE
                freqs_cos = self.freqs_cos[i]
                freqs_sin = self.freqs_sin[i]
                
                f_cos = orchard_core.Tensor(self.backend, [self.config['hidden_size'] // self.config['num_attention_heads'] // 2], orchard_core.DType.Float32)
                f_sin = orchard_core.Tensor(self.backend, [self.config['hidden_size'] // self.config['num_attention_heads'] // 2], orchard_core.DType.Float32)
                f_cos.copy_from_host(freqs_cos)
                f_sin.copy_from_host(freqs_sin)
                
                for layer in self.layers:
                    layer.forward(x, f_cos, f_sin, [b])
                    
            current_tokens.append(ids[-1])
            
        # Generation Loop
        print(f"Generating...")
        
        generated_sequences = [[] for _ in range(batch_size)]
        active_indices = list(range(batch_size))
        
        for step in range(max_tokens):
            if not active_indices:
                break
                
            # Prepare batch input
            curr_batch_size = len(active_indices)
            x_batch_host = np.zeros((curr_batch_size, self.config['hidden_size']), dtype=np.float32)
            
            head_dim = self.config['hidden_size'] // self.config['num_attention_heads']
            f_cos_host = np.zeros((curr_batch_size, head_dim // 2), dtype=np.float32)
            f_sin_host = np.zeros((curr_batch_size, head_dim // 2), dtype=np.float32)
            
            for i, b_idx in enumerate(active_indices):
                token_id = current_tokens[b_idx]
                x_batch_host[i] = self.embed_tokens[token_id]
                
                pos = len(batch_ids[b_idx]) + len(generated_sequences[b_idx])
                f_cos_host[i] = self.freqs_cos[pos]
                f_sin_host[i] = self.freqs_sin[pos]
                
            # Upload to GPU
            x_tensor = orchard_core.Tensor(self.backend, [curr_batch_size * self.config['hidden_size']], orchard_core.DType.Float32)
            x_tensor.copy_from_host(x_batch_host.flatten())
            
            f_cos_tensor = orchard_core.Tensor(self.backend, [curr_batch_size * (head_dim // 2)], orchard_core.DType.Float32)
            f_cos_tensor.copy_from_host(f_cos_host.flatten())
            
            f_sin_tensor = orchard_core.Tensor(self.backend, [curr_batch_size * (head_dim // 2)], orchard_core.DType.Float32)
            f_sin_tensor.copy_from_host(f_sin_host.flatten())
            
            # Forward
            for layer in self.layers:
                layer.forward(x_tensor, f_cos_tensor, f_sin_tensor, active_indices)
                
            # RMSNorm
            self.backend.run_rmsnorm(x_tensor, self.norm, x_tensor, self.config['rms_norm_eps'], self.config['hidden_size'], curr_batch_size)
            
            # Logits & Sampling
            logits_output = orchard_core.Tensor(self.backend, [curr_batch_size * self.config['vocab_size']], orchard_core.DType.Float32)
            
            if curr_batch_size == 1:
                self.backend.run_gemv_q4_0(self.output_w, self.output_s, x_tensor, logits_output, self.config['hidden_size'], self.config['vocab_size'])
            else:
                self.backend.run_gemm_q4_0(self.output_w, self.output_s, x_tensor, logits_output, self.config['hidden_size'], self.config['vocab_size'], curr_batch_size)
                
            logits_host = np.array(logits_output.copy_to_host(), copy=False).reshape(curr_batch_size, self.config['vocab_size'])
            
            # Sample
            next_active_indices = []
            
            for i, b_idx in enumerate(active_indices):
                logits = logits_host[i]
                next_token = self._sample(logits, temperature)
                
                generated_sequences[b_idx].append(next_token)
                current_tokens[b_idx] = next_token
                
                # Check EOS (assuming 2 is EOS for Llama 2)
                if next_token != 2: # EOS
                     next_active_indices.append(b_idx)
                     
            active_indices = next_active_indices
            
        # Decode
        results = []
        for seq in generated_sequences:
            results.append(self.tokenizer.decode(seq))
            
        return results




import os
import json
import glob
import numpy as np
from safetensors.numpy import save_file, load_file
from .quantizer import quantize_q4_0

def optimize_model(input_path: str, output_path: str):
    """
    Optimizes a model by pre-quantizing weights to Q4_0 and saving them.
    """
    if not os.path.exists(output_path):
        os.makedirs(output_path)
        
    print(f"Optimizing model from {input_path} to {output_path}...")
    
    # 1. Copy Config & Tokenizer
    # We just copy non-safetensors files
    for filename in os.listdir(input_path):
        if not filename.endswith(".safetensors") and not os.path.isdir(os.path.join(input_path, filename)):
            # Simple copy
            with open(os.path.join(input_path, filename), 'rb') as src, open(os.path.join(output_path, filename), 'wb') as dst:
                dst.write(src.read())
                
    # 2. Process Safetensors
    files = glob.glob(os.path.join(input_path, "*.safetensors"))
    
    for file_path in files:
        filename = os.path.basename(file_path)
        print(f"Processing {filename}...")
        
        tensors = load_file(file_path)
        new_tensors = {}
        
        for key, tensor in tensors.items():
            # Check if this is a weight we want to quantize
            # We quantize Linear layers in the transformer blocks and the output head
            # Keys usually look like: model.layers.0.self_attn.q_proj.weight
            
            should_quantize = False
            if "model.layers" in key and "weight" in key and "norm" not in key:
                # Attention and MLP weights
                should_quantize = True
            elif "lm_head.weight" in key:
                should_quantize = True
                
            if should_quantize:
                print(f"  Quantizing {key}...")
                # Ensure float32
                if tensor.dtype != np.float32:
                    tensor = tensor.astype(np.float32)
                    
                packed, scales = quantize_q4_0(tensor)
                
                new_tensors[key.replace(".weight", ".weight_packed")] = packed
                new_tensors[key.replace(".weight", ".weight_scales")] = scales
            else:
                # Keep as is (e.g. norms, embeddings)
                print(f"  Copying {key}...")
                new_tensors[key] = tensor
                
        # Save new file
        save_path = os.path.join(output_path, filename)
        save_file(new_tensors, save_path)
        
    print("Optimization complete.")



import sys
from setuptools import setup, Extension
import pybind11
import distutils.unixccompiler

# Monkeypatch to allow .mm files
if ".mm" not in distutils.unixccompiler.UnixCCompiler.src_extensions:
    distutils.unixccompiler.UnixCCompiler.src_extensions.append(".mm")

# Compiler flags
cflags = ["-std=c++17", "-Isrc"]
ldflags = ["-framework", "Metal", "-framework", "Foundation"]

# Define the extension
orchard_core = Extension(
    "orchard_core",
    sources=[
        "src/bindings.cpp",
        "src/platform/metal_backend.mm",
        "src/runtime/tensor.cpp",
        "src/runtime/model.cpp",
        "src/runtime/kv_cache.cpp"
    ],
    include_dirs=["src", pybind11.get_include()],
    extra_compile_args=cflags,
    extra_link_args=ldflags,
    language="c++"
)

setup(
    name="orchard-llm",
    version="0.1.1",
    description="High-performance LLM runtime for Apple Silicon",
    author="Orchard Team",
    packages=["orchard"],
    package_data={
        "orchard": ["kernels/*.metal"],
    },
    ext_modules=[orchard_core],
    install_requires=[
        "numpy",
        "safetensors",
        "tokenizers",
        "huggingface_hub"
    ],
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "orchard=orchard.cli:main",
        ],
    },
    python_requires=">=3.9",
)



import json
import os
import glob
import numpy as np
from safetensors import safe_open
from .quantizer import quantize_q4_0
import orchard_core

def load_weights(model, model_path):
    print(f"Loading model from {model_path}...")
    
    # 1. Load Config
    config_path = os.path.join(model_path, "config.json")
    with open(config_path, "r") as f:
        config = json.load(f)
    
    model.config = config
    n_layers = config['num_hidden_layers']
    hidden_dim = config['hidden_size']
    
    # Initialize layers
    from .model import LlamaLayer
    model.layers = [LlamaLayer(model.backend, config, i) for i in range(n_layers)]
    
    # 2. Find Safetensors files
    files = glob.glob(os.path.join(model_path, "*.safetensors"))
    if not files:
        raise FileNotFoundError("No .safetensors files found")
    
    # 3. Load Weights
    # We iterate over files and load what we find
    for file in files:
        print(f"Processing {os.path.basename(file)}...")
        with safe_open(file, framework="numpy", device="cpu") as f:
            keys = f.keys()
            for key in keys:
                if key.endswith("_scales"):
                    continue # Handled by _packed
                
                # Parse key to find layer and type
                # e.g. model.layers.0.self_attn.q_proj.weight
                parts = key.split('.')
                is_packed = key.endswith("_packed")
                
                if parts[0] == "model" and parts[1] == "layers":
                    layer_idx = int(parts[2])
                    layer = model.layers[layer_idx]
                    module = parts[3] # self_attn or mlp or input_layernorm
                    
                    if module == "self_attn" or module == "mlp":
                        proj = parts[4]
                        target_w, target_s = None, None
                        
                        if module == "self_attn":
                            if proj == "q_proj": target_w, target_s = "wq_w", "wq_s"
                            elif proj == "k_proj": target_w, target_s = "wk_w", "wk_s"
                            elif proj == "v_proj": target_w, target_s = "wv_w", "wv_s"
                            elif proj == "o_proj": target_w, target_s = "wo_w", "wo_s"
                        elif module == "mlp":
                            if proj == "gate_proj": target_w, target_s = "w1_w", "w1_s"
                            elif proj == "up_proj": target_w, target_s = "w3_w", "w3_s"
                            elif proj == "down_proj": target_w, target_s = "w2_w", "w2_s"
                            
                        if target_w:
                            if is_packed:
                                packed = f.get_tensor(key)
                                scales = f.get_tensor(key.replace("_packed", "_scales"))
                                w, s = _upload_quantized(model.backend, packed, scales)
                            else:
                                tensor = f.get_tensor(key)
                                w, s = _quantize_and_upload(model.backend, tensor)
                            
                            setattr(layer, target_w, w)
                            setattr(layer, target_s, s)
                            
                    elif module == "input_layernorm":
                        # FP32 upload
                        layer.attn_norm = _upload_fp32(model.backend, f.get_tensor(key))
                    elif module == "post_attention_layernorm":
                        layer.ffn_norm = _upload_fp32(model.backend, f.get_tensor(key))
                        
                elif parts[0] == "model" and parts[1] == "norm":
                    # Final norm
                    tensor = f.get_tensor(key)
                    model.norm = _upload_fp32(model.backend, tensor)
                
                elif parts[0] == "model" and parts[1] == "embed_tokens":
                    # Embeddings
                    # Keep on CPU as numpy array for v0.1
                    tensor = f.get_tensor(key)
                    if tensor.dtype != np.float32:
                        tensor = tensor.astype(np.float32)
                    model.embed_tokens = tensor
                    
                elif parts[0] == "lm_head":
                    # Output head
                    if is_packed:
                        packed = f.get_tensor(key)
                        scales = f.get_tensor(key.replace("_packed", "_scales"))
                        w, s = _upload_quantized(model.backend, packed, scales)
                    else:
                        tensor = f.get_tensor(key)
                        w, s = _quantize_and_upload(model.backend, tensor)
                    model.output_w = w; model.output_s = s

def _upload_quantized(backend, packed, scales):
    rows, cols_packed = packed.shape
    t_w = orchard_core.Tensor(backend, [rows, cols_packed], orchard_core.DType.Int8)
    t_w.copy_from_host(packed)
    
    rows, cols_scales = scales.shape
    t_s = orchard_core.Tensor(backend, [rows, cols_scales], orchard_core.DType.Float16)
    t_s.copy_from_host(scales)
    
    return t_w, t_s

def _quantize_and_upload(backend, tensor_np):
    # tensor_np is (Out, In)
    # Quantize
    # Ensure float32 for quantization precision
    if tensor_np.dtype != np.float32:
        tensor_np = tensor_np.astype(np.float32)
        
    packed, scales = quantize_q4_0(tensor_np)
    
    # Upload
    # Packed: (N, K/2) uint8
    # Scales: (N, K/32) float16
    
    rows, cols_packed = packed.shape
    t_w = orchard_core.Tensor(backend, [rows, cols_packed], orchard_core.DType.Int8)
    t_w.copy_from_host(packed)
    
    rows, cols_scales = scales.shape
    t_s = orchard_core.Tensor(backend, [rows, cols_scales], orchard_core.DType.Float16)
    t_s.copy_from_host(scales)
    
    return t_w, t_s

def _upload_fp32(backend, tensor_np):
    # Convert to float32 if not already
    if tensor_np.dtype != np.float32:
        tensor_np = tensor_np.astype(np.float32)
        
    shape = tensor_np.shape
    t = orchard_core.Tensor(backend, list(shape), orchard_core.DType.Float32)
    t.copy_from_host(tensor_np)
    return t

def load_lora_adapter(model, adapter_path):
    print(f"Loading LoRA adapter from {adapter_path}...")
    
    # 1. Load Config
    config_path = os.path.join(adapter_path, "adapter_config.json")
    with open(config_path, "r") as f:
        config = json.load(f)
        
    r = config.get("r", 8)
    alpha = config.get("lora_alpha", 16)
    scaling = alpha / r
    
    # 2. Find Weights
    files = glob.glob(os.path.join(adapter_path, "*.safetensors"))
    if not files:
        # Try bin
        files = glob.glob(os.path.join(adapter_path, "*.bin"))
        if not files:
            raise FileNotFoundError("No adapter weights found")
            
    # 3. Load Weights
    from .model import LoraAdapter
    
    for file in files:
        print(f"Processing adapter {os.path.basename(file)}...")
        # Use safe_open or torch.load depending on file type
        # Assuming safetensors for now as it's standard
        if file.endswith(".safetensors"):
            with safe_open(file, framework="numpy", device="cpu") as f:
                keys = f.keys()
                for key in keys:
                    # Parse key
                    # base_model.model.model.layers.0.self_attn.q_proj.lora_A.weight
                    parts = key.split('.')
                    
                    # Find layer index
                    try:
                        if "layers" in parts:
                            layer_idx_pos = parts.index("layers") + 1
                            layer_idx = int(parts[layer_idx_pos])
                        else:
                            continue
                    except (ValueError, IndexError):
                        continue
                        
                    if layer_idx >= len(model.layers):
                        continue
                        
                    layer = model.layers[layer_idx]
                    
                    # Find module
                    # self_attn.q_proj or mlp.gate_proj
                    try:
                        module_part = parts[layer_idx_pos + 1]
                        proj_part = parts[layer_idx_pos + 2]
                    except IndexError:
                        continue
                    
                    target_module = None
                    if module_part == "self_attn":
                        if proj_part == "q_proj": target_module = "wq"
                        elif proj_part == "k_proj": target_module = "wk"
                        elif proj_part == "v_proj": target_module = "wv"
                        elif proj_part == "o_proj": target_module = "wo"
                    elif module_part == "mlp":
                        if proj_part == "gate_proj": target_module = "w1"
                        elif proj_part == "up_proj": target_module = "w3"
                        elif proj_part == "down_proj": target_module = "w2"
                        
                    if not target_module:
                        continue
                        
                    # Check if A or B
                    is_A = "lora_A" in parts
                    is_B = "lora_B" in parts
                    
                    if not (is_A or is_B):
                        continue
                        
                    # Get or create adapter
                    if target_module not in layer.lora_adapters:
                        layer.lora_adapters[target_module] = LoraAdapter(model.backend, r, alpha, scaling)
                    
                    adapter = layer.lora_adapters[target_module]
                    tensor = f.get_tensor(key)
                    
                    # Upload
                    # A: [r, In] -> Transpose to [In, r]
                    # B: [Out, r] -> Transpose to [r, Out]
                    
                    if is_A:
                        # tensor is [r, In]
                        # We want [In, r]
                        tensor_t = tensor.T.astype(np.float32)
                        t = orchard_core.Tensor(model.backend, list(tensor_t.shape), orchard_core.DType.Float32)
                        t.copy_from_host(tensor_t)
                        adapter.lora_A = t
                    elif is_B:
                        # tensor is [Out, r]
                        # We want [r, Out]
                        # Also apply scaling here!
                        tensor_scaled = tensor.T.astype(np.float32) * scaling
                        t = orchard_core.Tensor(model.backend, list(tensor_scaled.shape), orchard_core.DType.Float32)
                        t.copy_from_host(tensor_scaled)
                        adapter.lora_B = t

